#
# Copyright 2018 Russell Smiley
#
# This file is part of package-manifest.
#
# package-manifest is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# package-manifest is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with package-manifest.  If not, see <http://www.gnu.org/licenses/>.
#

import unittest

from ..globalName import Include as GlobalInclude
from ..localName import Exclude as LocalExclude
from ..tree import Prune

from ..manifest import \
    Manifest, \
    YamlNotFound


class TestManifestYaml( unittest.TestCase ) :
    def testNoneYamlRaises( self ) :
        expectedRootDirectory = '.'
        with self.assertRaisesRegex( YamlNotFound, '^No package-manifest YAML data present' ) :
            Manifest.from_yamlData( None, expectedRootDirectory )


    def testEmptyDictYamlRaises( self ) :
        expectedRootDirectory = '.'
        with self.assertRaisesRegex( YamlNotFound, '^No package-manifest YAML data present' ) :
            Manifest.from_yamlData( dict(), expectedRootDirectory )


    def testEmptyListYamlRaises( self ) :
        expectedRootDirectory = '.'
        with self.assertRaisesRegex( YamlNotFound, '^No package-manifest YAML data present' ) :
            Manifest.from_yamlData( list(), expectedRootDirectory )


    def testDecodeOperationsList( self ) :
        expectedRootDirectory = '.'

        inputYamlData = [
            { 'global-include' :
                  { 'files' :
                        [ '*.h' ] }, },
            { 'exclude' :
                  { 'files' :
                        [ 'LICENSE', 'README' ] }, },
            { 'prune' :
                  { 'directory' : 'bin' }, },
        ]

        manifestUnderTest = Manifest.from_yamlData( inputYamlData, expectedRootDirectory )

        self.assertEqual( expectedRootDirectory, manifestUnderTest.pathSet.rootDirectory )

        self.assertEqual( 4, len( manifestUnderTest.operations ) )

        self.assertTrue( isinstance( manifestUnderTest.operations[ 0 ], GlobalInclude ) )
        self.assertTrue( isinstance( manifestUnderTest.operations[ 1 ], LocalExclude ) )
        self.assertTrue( isinstance( manifestUnderTest.operations[ 2 ], LocalExclude ) )
        self.assertTrue( isinstance( manifestUnderTest.operations[ 3 ], Prune ) )


if __name__ == '__main__' :
    unittest.main()



#
# Copyright 2017 Russell Smiley
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import unittest

from ..path import makeRelativePaths


class TestMakeRelativePaths( unittest.TestCase ) :
    def testGoodValues( self ) :
        rootPath = '/some/path'
        inputList = [
            '/some/path/file1',
            '/some/path/file2',
            '/some/path/subdir/file3',
        ]
        expectedPaths = [
            'file1',
            'file2',
            'subdir/file3',
        ]

        actualPaths = makeRelativePaths( inputList, rootPath )

        self.assertEqual( actualPaths, expectedPaths )


if __name__ == '__main__' :
    unittest.main( )



#
# Copyright 2018 Russell Smiley
#
# This file is part of package-manifest.
#
# package-manifest is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# package-manifest is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with package-manifest.  If not, see <http://www.gnu.org/licenses/>.
#

import fnmatch
import re

from ..path.path import PathOperation
from ..yamlParser import \
    Files, \
    YamlNotFound


def applyGlobalInclusions( pathNames, inclusions ) :
    """
    Apply unix glob wildcard expressions to select items from an iterable of file names (basename). The wildcard is applied
    "globally".

    eg.
      "*.py" implies "select *.py files anywhere in the root directory and it's subdirectories, recursively".

    :param pathNames: path names to be filtered expressed relative to the root directory of the package-manifest.
    :param inclusions: include patterns

    :return: set of path names after inclusions have been applied.
    """
    previousNames = set()
    for thisInclude in inclusions :
        for thisName in pathNames :
            regularExpression = fnmatch.translate( thisInclude ) + '$'
            if re.search( regularExpression, thisName ) is not None :
                previousNames.add( thisName )

    return previousNames


class Include( PathOperation ) :
    __yamlKey = 'global-include'


    def __init__( self, includePattern ) :
        self.includePattern = includePattern


    def apply( self, originalPathNames, currentPathNames ) :
        assert all( [ x in originalPathNames for x in currentPathNames ] )

        includedNames = applyGlobalInclusions( originalPathNames, [ self.includePattern ] )
        modifiedNames = currentPathNames | includedNames

        return modifiedNames


    @classmethod
    def from_yamlData( cls, yamlData ) :
        if yamlData is None :
            raise YamlNotFound( 'yamlData is None' )
        elif Include.__yamlKey not in yamlData :
            raise YamlNotFound( '{1} not found in yaml data, {0}'.format( yamlData, Include.__yamlKey ) )
        else :
            try :
                filePatterns = Files.from_yamlData( yamlData[ Include.__yamlKey ] )

                result = list()
                for thisPattern in filePatterns.patterns :
                    result.append( cls( thisPattern ) )
            except YamlNotFound as e :
                raise YamlNotFound( '{1} must use files directive, {0}'.format( yamlData, Include.__yamlKey ) ) from e

        return result



#
# Copyright 2018 Russell Smiley
#
# This file is part of package-manifest.
#
# package-manifest is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# package-manifest is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with package-manifest.  If not, see <http://www.gnu.org/licenses/>.
#

import unittest

from ..exclude import Exclude


class TestLocalNameExclude( unittest.TestCase ) :
    def testSimpleEmptyExclude( self ) :
        exclusion = 't*'
        inputList = { 'one', 'two', 'three', 'four' }
        expectedList = set()

        excludeUnderTest = Exclude( exclusion )

        actualList = excludeUnderTest.apply( inputList, set() )

        self.assertEqual( actualList, expectedList )


    def testSimpleNonemptyExclude( self ) :
        exclusion = 't*'
        inputList = { 'one', 'two', 'three', 'four' }
        expectedList = { 'four' }

        excludeUnderTest = Exclude( exclusion )

        actualList = excludeUnderTest.apply( inputList, { 'two', 'four' } )

        self.assertEqual( actualList, expectedList )


if __name__ == '__main__' :
    unittest.main()



#
# Copyright 2018 Russell Smiley
#
# This file is part of package-manifest.
#
# package-manifest is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# package-manifest is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with package-manifest.  If not, see <http://www.gnu.org/licenses/>.
#

import os
import unittest

from ..manifest import loadYamlData


class TestLoadYaml( unittest.TestCase ) :
    def setUp( self ) :
        thisDir = os.path.dirname( __file__ )

        self.dataDir = os.path.join( thisDir, 'data' )


    def test_something( self ) :
        yamlFilePath = os.path.join( self.dataDir, 'sampleManifest.yml' )

        yamlData = loadYamlData( yamlFilePath )

        self.assertTrue( isinstance( yamlData, list ) )
        self.assertEqual( 8, len( yamlData ) )


if __name__ == '__main__' :
    unittest.main()



#
# Copyright 2018 Russell Smiley
#
# This file is part of package-manifest.
#
# package-manifest is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# package-manifest is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with package-manifest.  If not, see <http://www.gnu.org/licenses/>.
#

import unittest

from ..exclude import \
    Exclude, \
    YamlNotFound


class TestGlobalNameYamlExclude( unittest.TestCase ) :
    def testNoneNotFound( self ) :
        yamlData = None

        with self.assertRaisesRegex( YamlNotFound, '^yamlData is None' ) :
            Exclude.from_yamlData( yamlData )


    def testExcludeNotFound( self ) :
        yamlData = { 'files' : [ 'one' ] }

        with self.assertRaisesRegex( YamlNotFound, '^global-exclude not found in yaml data' ) :
            Exclude.from_yamlData( yamlData )


    def testExcludeFilesNotFound( self ) :
        yamlData = { 'global-exclude' : { 'directory' : 'one' } }

        with self.assertRaisesRegex( YamlNotFound, '^global-exclude must use files directive' ) :
            Exclude.from_yamlData( yamlData )


    def testExcludeOk( self ) :
        expectedPatterns = [ 'one', 't*' ]
        yamlData = { 'global-exclude' : { 'files' : expectedPatterns } }

        excludesUnderTest = Exclude.from_yamlData( yamlData )

        self.assertTrue( isinstance( excludesUnderTest, list ) )
        self.assertEqual( len( expectedPatterns ), len( excludesUnderTest ) )
        self.assertEqual( expectedPatterns[ 0 ], excludesUnderTest[ 0 ].excludePattern )
        self.assertEqual( expectedPatterns[ 1 ], excludesUnderTest[ 1 ].excludePattern )


if __name__ == '__main__' :
    unittest.main()



#
# Copyright 2018 Russell Smiley
#
# This file is part of package-manifest.
#
# package-manifest is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# package-manifest is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with package-manifest.  If not, see <http://www.gnu.org/licenses/>.
#

from .exception import YamlNotFound


class Directory :
    def __init__( self ) :
        self.path = None


    @classmethod
    def from_yamlData( cls, yamlData ) :
        if yamlData is None :
            raise YamlNotFound( 'yamlData is None' )
        elif 'directory' not in yamlData :
            raise YamlNotFound( 'directory not specified in yaml, {0}'.format( yamlData ) )
        else :
            thisDirectory = cls()

            thisDirectory.path = yamlData[ 'directory' ]

        return thisDirectory



#
# Copyright 2018 Russell Smiley
#
# This file is part of package-manifest.
#
# package-manifest is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# package-manifest is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with package-manifest.  If not, see <http://www.gnu.org/licenses/>.
#

from ..path.path import PathOperation
from ..yamlParser import \
    Files, \
    YamlNotFound

from .include import applyInclusions


def applyExclusions( currentNames, exclusions ) :
    includedNames = applyInclusions( currentNames, exclusions )

    retainedPaths = set( currentNames ) - includedNames

    return retainedPaths


class Exclude( PathOperation ) :
    __yamlKey = 'exclude'


    def __init__( self, excludePattern ) :
        self.excludePattern = excludePattern


    def apply( self, originalPathNames, currentPathNames ) :
        assert all( [ x in originalPathNames for x in currentPathNames ] )

        retainedPaths = applyExclusions( originalPathNames, [ self.excludePattern ] )
        modifiedNames = currentPathNames & retainedPaths

        return modifiedNames


    @classmethod
    def from_yamlData( cls, yamlData ) :
        if yamlData is None :
            raise YamlNotFound( 'yamlData is None' )
        elif Exclude.__yamlKey not in yamlData :
            raise YamlNotFound( '{1} not found in yaml data, {0}'.format( yamlData, Exclude.__yamlKey ) )
        else :
            try :
                filePatterns = Files.from_yamlData( yamlData[ Exclude.__yamlKey ] )

                result = list()
                for thisPattern in filePatterns.patterns :
                    result.append( cls( thisPattern ) )
            except YamlNotFound as e :
                raise YamlNotFound( '{1} must use files directive, {0}'.format( yamlData, Exclude.__yamlKey ) ) from e

        return result



#
# Copyright 2018 Russell Smiley
#
# This file is part of package-manifest.
#
# package-manifest is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# package-manifest is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with package-manifest.  If not, see <http://www.gnu.org/licenses/>.
#

import os
import unittest

from ..path import prependDirectory


class TestPrependDirectory( unittest.TestCase ) :
    def test_something( self ) :
        directory = os.path.join( 'some', 'base' )
        inputPaths = {
            os.path.join( 'one', 'two' ),
            'three',
        }

        expectedValue = {
            os.path.join( *[ 'some', 'base', 'one', 'two' ] ),
            os.path.join( *[ 'some', 'base', 'three' ] ),
        }

        actualValue = set( prependDirectory( directory, inputPaths ) )

        self.assertEqual( expectedValue, actualValue )


if __name__ == '__main__' :
    unittest.main()



#
# Copyright 2017 Russell Smiley
#
# This file is part of package-manifest.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import unittest

from ..include import applyInclusions


class TestLocalApplyInclusions( unittest.TestCase ) :
    def functionUnderTest( self, inputList, inclusions ) :
        return applyInclusions( inputList, inclusions )


    def testSimpleInclude( self ) :
        inclusions = [ 't*' ]
        inputList = [ 'one', 'two', 'three', 'four' ]
        expectedList = { 'two', 'three' }

        actualList = set( self.functionUnderTest( inputList, inclusions ) )

        self.assertEqual( actualList, expectedList )


    def testIncludeTwoWildcards( self ) :
        inclusions = [ 'tw*', 'th*' ]
        inputList = [ 'one', 'two', 'three', 'four' ]
        expectedList = { 'two', 'three' }

        actualList = set( self.functionUnderTest( inputList, inclusions ) )

        self.assertEqual( actualList, expectedList )


    def testIncludePathWildcards( self ) :
        inclusions = [ 'one/tw*' ]
        inputList = [ 'one', 'one/twill', 'two', 'three', 'one/two', 'four' ]
        expectedList = { 'one/two', 'one/twill' }

        actualList = set( self.functionUnderTest( inputList, inclusions ) )

        self.assertEqual( actualList, expectedList )


    def testSubdirInclude( self ) :
        inclusions = [ 't*' ]
        inputList = [ 'one', 'two', 'three', 'four', 'subdir/one', 'subdir/two' ]
        expectedList = { 'two', 'three' }

        actualList = set( self.functionUnderTest( inputList, inclusions ) )

        self.assertEqual( actualList, expectedList )


if __name__ == '__main__' :
    unittest.main()



#
# Copyright 2018 Russell Smiley
#
# This file is part of package-manifest.
#
# package-manifest is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# package-manifest is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with package-manifest.  If not, see <http://www.gnu.org/licenses/>.
#

import os
import unittest

from ..localName import Include as LocalInclude
from ..globalName import Exclude as GlobalExclude
from ..globalName import Include as GlobalInclude

from ..manifest import Manifest


class MockPathSet :
    def __init__( self ) :
        self.pathNames = set()
        self.rootDirectory = '.'


class TestManifest( unittest.TestCase ) :
    def setUp( self ) :
        self.pathSet = MockPathSet()

        self.pathSet.pathNames = {
            os.path.join( 'LICENSE' ),
            os.path.join( 'README' ),
            os.path.join( 'include', 'main.h' ),
            os.path.join( 'include', 'other.h' ),
        }


    def testInitialNames( self ) :
        self.testManifest = Manifest( pathSet = self.pathSet )
        self.testManifest.operations.append( LocalInclude( 'LICENSE' ) )

        actualNames = self.testManifest.apply()

        self.assertIn( 'LICENSE', actualNames )

        self.assertNotIn( os.path.join( 'README' ), actualNames )
        self.assertNotIn( os.path.join( 'include', 'main.h' ), actualNames )
        self.assertNotIn( os.path.join( 'include', 'other.h' ), actualNames )


    def testMultipleOperations( self ) :
        self.testManifest = Manifest( pathSet = self.pathSet )
        self.testManifest.operations.append( GlobalInclude( '*.h' ) )
        self.testManifest.operations.append( GlobalExclude( os.path.join( 'include', 'other.h' ) ) )

        actualNames = self.testManifest.apply()

        self.assertNotIn( os.path.join( 'LICENSE' ), actualNames )
        self.assertNotIn( os.path.join( 'README' ), actualNames )

        self.assertIn( os.path.join( 'include', 'main.h' ), actualNames )


if __name__ == '__main__' :
    unittest.main()



#
# Copyright 2018 Russell Smiley
#
# This file is part of package-manifest.
#
# package-manifest is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# package-manifest is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with package-manifest.  If not, see <http://www.gnu.org/licenses/>.
#

import os

from .path import makeRelativePaths


class PathSet :
    def __init__( self, rootDirectory ) :
        self.pathNames = set()
        self.rootDirectory = rootDirectory

        if not os.path.isdir( self.rootDirectory ) :
            raise RuntimeError( 'Manifest root directory does not exist, ' + repr( self.rootDirectory ) )

        self.__buildOriginalFiles()


    def __buildOriginalFiles( self ) :
        # Recursively acquire all files/folders in the root directory.
        for root, directories, files in os.walk( self.rootDirectory ) :
            for thisDir in directories :
                self.pathNames.add( os.path.join( root, thisDir ) )

            for thisFile in files :
                self.pathNames.add( os.path.join( root, thisFile ) )

        self.__makeFilesRelative()


    def __makeFilesRelative( self ) :
        self.pathNames = makeRelativePaths( self.pathNames, self.rootDirectory )



#
# Copyright 2018 Russell Smiley
#
# This file is part of package-manifest.
#
# package-manifest is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# package-manifest is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with package-manifest.  If not, see <http://www.gnu.org/licenses/>.
#

import os
import unittest

from ..include import Include


class TestRecursiveInclude( unittest.TestCase ) :
    def testGlobalInclude( self ) :
        directory = '.'
        inputNames = {
            os.path.join( 'one' ),
            os.path.join( 'subdir', 'two' ),
            os.path.join( 'sub', 'subdir', 'three' ),
            os.path.join( 'subdir', 'four' ),
            os.path.join( 'twice' ),
        }
        inclusion = 't*'
        expectedValue = {
            os.path.join( 'subdir', 'two' ),
            os.path.join( 'twice' ),
            os.path.join( 'sub', 'subdir', 'three' ),
        }

        includeUnderTest = Include( directory, inclusion )

        actualValue = includeUnderTest.apply( inputNames, set() )

        self.assertEqual( { os.path.normpath( x ) for x in expectedValue }, actualValue )


if __name__ == '__main__' :
    unittest.main()



#
# Copyright 2018 Russell Smiley
#
# This file is part of package-manifest.
#
# package-manifest is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# package-manifest is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with package-manifest.  If not, see <http://www.gnu.org/licenses/>.
#

from .exception import YamlNotFound

from .directory import Directory
from .files import Files



#
# Copyright 2018 Russell Smiley
#
# This file is part of package-manifest.
#
# package-manifest is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# package-manifest is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with package-manifest.  If not, see <http://www.gnu.org/licenses/>.
#

import abc
import os.path
import re


def filterNotInDirectory( pathNames, directory ) :
    """
    Remove paths from an iterable of path names that are not recursively "inside" the specified directory.

    :param pathNames: iterable of paths.
    :param directory: directory path.

    :return: list of paths inside `directory`.
    """
    relativeNames = makeRelativePaths( pathNames, directory )

    currentDirNames = list()
    for thisName in relativeNames :
        if re.match( r'^\.\.', thisName ) is None :
            currentDirNames.append( thisName )

    return prependDirectory( directory, currentDirNames )


def makeRelativePaths( paths, pathRoot ) :
    """
    Turn an iterable of paths and express them as relative to the specified root directory path using `os.path.relpath`.

    Assumes that `paths` are valid for use with `os.path.relpath` against `pathRoot`.

    :param paths: iterable of paths
    :param pathRoot: directory path stem of paths

    :return: list of resulting relative paths.
    """
    relativePaths = [ os.path.relpath( x, pathRoot ) for x in paths ]
    return relativePaths


def prependDirectory( directory, names ) :
    """
    Prepend directory path to an iterable of path names.

    :param directory: directory path to be prepended to names
    :param names: iterable of names

    :return: list of prepended names.
    """
    return [ os.path.normpath( os.path.join( directory, x ) ) for x in names ]


class PathOperation( metaclass = abc.ABCMeta ) :
    @abc.abstractmethod
    def apply( self, originalPathNames, currentPathNames ) :
        """
        Apply a path operation to a set of path names.

        :param originalPathNames: The original list of all path names
        :param currentPathNames: The current list of path names, perhaps as processed by other (previous) path operations.

        :return: Modified currentPathNames combining this path operation with the effects of previous path operations.
        """
        pass



#
# Copyright 2018 Russell Smiley
#
# This file is part of package-manifest.
#
# package-manifest is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# package-manifest is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with package-manifest.  If not, see <http://www.gnu.org/licenses/>.
#

import os
import unittest

from ..set import PathSet


class TestPathSet( unittest.TestCase ) :
    @classmethod
    def setUpClass( cls ) :
        cls.thisDir = os.path.dirname( __file__ )


    def testAcquirePaths( self ) :
        p = PathSet( self.thisDir )

        self.assertIn( os.path.join( 'testPathSet.py' ), p.pathNames )
        self.assertIn( os.path.join( 'testPrependDirectory.py' ), p.pathNames )


if __name__ == '__main__' :
    unittest.main()



#
# Copyright 2018 Russell Smiley
#
# This file is part of package-manifest.
#
# package-manifest is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# package-manifest is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with package-manifest.  If not, see <http://www.gnu.org/licenses/>.
#

import unittest

from ..directory import \
    Directory, \
    YamlNotFound


class TestYamlDirectory( unittest.TestCase ) :
    def testNoneNotFound( self ) :
        yamlData = None

        with self.assertRaisesRegex( YamlNotFound, '^yamlData is None' ) :
            Directory.from_yamlData( yamlData )


    def testNotDirectoryNotFound( self ) :
        yamlData = { 'files' : [ 'one', 'two' ] }

        with self.assertRaisesRegex( YamlNotFound, '^directory not specified in yaml' ) :
            Directory.from_yamlData( yamlData )


    def testPathOk( self ) :
        expectedPath = 'path'
        yamlData = { 'directory' : expectedPath }

        directoryUnderTest = Directory.from_yamlData( yamlData )

        self.assertEqual( expectedPath, directoryUnderTest.path )


if __name__ == '__main__' :
    unittest.main()



#
# Copyright 2018 Russell Smiley
#
# This file is part of package-manifest.
#
# package-manifest is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# package-manifest is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with package-manifest.  If not, see <http://www.gnu.org/licenses/>.
#

from ..path.path import PathOperation

from ..yamlParser import \
    Directory, \
    YamlNotFound

from .graft import graftPaths


def prunePaths( pathNames, directory ) :
    """
    Prune (remove) a directory and it's files, recursively, from the set of input paths.

    :param pathNames: Iterable of paths from which to select files in directory.
    :param directory: Directory to select pruned files from relative to the package-manifest root directory.

    :return: set of filtered files
    """
    graftedPaths = graftPaths( pathNames, directory )

    prunedPaths = set( pathNames ) - graftedPaths

    return prunedPaths


class Prune( PathOperation ) :
    __yamlKey = 'prune'


    def __init__( self, directory ) :
        self.directory = directory


    def apply( self, originalPathNames, currentPathNames ) :
        assert all( [ x in originalPathNames for x in currentPathNames ] )

        retainedPaths = prunePaths( originalPathNames, self.directory )
        modifiedPaths = currentPathNames & retainedPaths

        return modifiedPaths


    @classmethod
    def from_yamlData( cls, yamlData ) :
        if yamlData is None :
            raise YamlNotFound( 'yamlData is None' )
        elif Prune.__yamlKey not in yamlData :
            raise YamlNotFound( '{1} not found in yaml data, {0}'.format( yamlData, Prune.__yamlKey ) )
        else :
            try :
                directory = Directory.from_yamlData( yamlData[ Prune.__yamlKey ] )

                # Need to return a list even though there is only one element in order to maintain compatibility with
                # the local, global & recursive operations.
                result = [ cls( directory.path ) ]
            except YamlNotFound as e :
                raise YamlNotFound(
                    '{1} must use directory directive, {0}'.format( yamlData, Prune.__yamlKey ) ) from e

        return result



#
# Copyright 2018 Russell Smiley
#
# This file is part of package-manifest.
#
# package-manifest is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# package-manifest is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with package-manifest.  If not, see <http://www.gnu.org/licenses/>.
#

import logging
import yaml

from .globalName import Include as GlobalInclude
from .globalName import Exclude as GlobalExclude
from .localName import Include as LocalInclude
from .localName import Exclude as LocalExclude
from .recursive import Include as RecursiveInclude
from .recursive import Exclude as RecursiveExclude
from .tree import \
    Graft, \
    Prune

from .path import PathSet

from .yamlParser import YamlNotFound


log = logging.getLogger( __name__ )


class Manifest :
    __operationYamlParsers = [
        GlobalExclude.from_yamlData,
        GlobalInclude.from_yamlData,
        LocalExclude.from_yamlData,
        LocalInclude.from_yamlData,
        RecursiveExclude.from_yamlData,
        RecursiveInclude.from_yamlData,
        Graft.from_yamlData,
        Prune.from_yamlData,
    ]


    def __init__( self,
                  rootDirectory = None,
                  pathSet = None ) :
        assert (rootDirectory is not None) ^ (pathSet is not None)

        if rootDirectory is not None :
            self.pathSet = PathSet( rootDirectory )
        elif pathSet is not None :
            assert hasattr( pathSet, 'pathNames' )

            self.pathSet = pathSet

        self.operations = list()


    def apply( self ) :
        processedFiles = set()
        for thisOperation in self.operations :
            processedFiles = thisOperation.apply( self.pathSet.pathNames, processedFiles )

        return processedFiles


    @classmethod
    def from_yamlData( cls, yamlData, rootDirectory ) :
        if not yamlData :
            raise YamlNotFound( 'No package-manifest YAML data present' )

        thisManifest = cls( rootDirectory = rootDirectory )

        for thisElement in yamlData :
            isValidOperationYaml = False
            for thisParser in cls.__operationYamlParsers :
                try :
                    operation = thisParser( thisElement )
                    thisManifest.operations += operation

                    isValidOperationYaml = True
                except YamlNotFound as e :
                    isValidOperationYaml |= False

                    log.debug( repr( e ) )

        if not isValidOperationYaml :
            raise YamlNotFound( 'No valid operations found in YAML element, {0}'.format( thisElement ) )

        return thisManifest


    @classmethod
    def from_yamlFile( cls, yamlFilePath, rootDirectory ) :
        yamlData = loadYamlData( yamlFilePath )
        thisManifest = cls.from_yamlData( yamlData, rootDirectory )

        return thisManifest


def loadYamlData( yamlFilePath ) :
    with open( yamlFilePath, 'r' ) as yamlFile :
        yamlData = yaml.load( yamlFile )

    return yamlData



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['packageManifest',
 'packageManifest.globalName',
 'packageManifest.globalName.tests',
 'packageManifest.localName',
 'packageManifest.localName.tests',
 'packageManifest.path',
 'packageManifest.path.tests',
 'packageManifest.recursive',
 'packageManifest.recursive.tests',
 'packageManifest.tests',
 'packageManifest.tree',
 'packageManifest.tree.tests',
 'packageManifest.yamlParser',
 'packageManifest.yamlParser.tests']

package_data = \
{'': ['*'], 'packageManifest.tests': ['data/*']}

install_requires = \
['pyyaml']

setup(name='packageManifest',
      version='0.2.0',
      description='A YAML based file package-manifest framework for defining packages.',
      author='Russell Smiley',
      author_email='im.russell.smiley@gmail.com',
      url='https://gitlab.com/blueskyjunkie/package-manifest',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      python_requires='>=3',
     )



#
# Copyright 2018 Russell Smiley
#
# This file is part of package-manifest.
#
# package-manifest is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# package-manifest is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with package-manifest.  If not, see <http://www.gnu.org/licenses/>.
#

from ..path.path import PathOperation
from ..yamlParser import \
    Directory, \
    Files, \
    YamlNotFound

from .include import applyRecursiveInclusions


def applyRecursiveExclusions( pathNames, directory, exclusions ) :
    """
    Apply an iterable of unix glob wildcard exclusion patterns recursively to an iterable of pathnames from the
    specified directory.

    Assume that path names, directory and exclusions are expressed relative to the same root directory.

    :param pathNames: path names to be filtered expressed relative to the root directory of the package-manifest.
    :param directory: directory to apply patterns to expressed relative to the root directory of the package-manifest.
    :param exclusions: exclude patterns

    :return: set of path names after exclusions have been applied.
    """
    includedPaths = applyRecursiveInclusions( pathNames, directory, exclusions )

    retainedPaths = set( pathNames ) - includedPaths

    return retainedPaths


class Exclude( PathOperation ) :
    __yamlKey = 'recursive-exclude'


    def __init__( self, directory, excludePattern ) :
        self.directory = directory
        self.excludePattern = excludePattern


    def apply( self, originalPathNames, currentPathNames ) :
        assert all( [ x in originalPathNames for x in currentPathNames ] )

        retainedPaths = applyRecursiveExclusions(originalPathNames, self.directory, [ self.excludePattern ])
        modifiedPaths = currentPathNames & retainedPaths

        return modifiedPaths


    @classmethod
    def from_yamlData( cls, yamlData ) :
        if yamlData is None :
            raise YamlNotFound( 'yamlData is None' )
        elif Exclude.__yamlKey not in yamlData :
            raise YamlNotFound( '{1} not found in yaml data, {0}'.format( yamlData, Exclude.__yamlKey ) )
        else :
            try :
                directory = Directory.from_yamlData( yamlData[ Exclude.__yamlKey ] )
                filePatterns = Files.from_yamlData( yamlData[ Exclude.__yamlKey ] )

                result = list()
                for thisPattern in filePatterns.patterns :
                    result.append( cls( directory.path, thisPattern ) )
            except YamlNotFound as e :
                raise YamlNotFound(
                    '{1} must use files and directory directives, {0}'.format( yamlData, Exclude.__yamlKey ) ) from e

        return result
